@@ -115,15 +115,6 @@ static const char *ddcSymbols[] = {
     NULL
 };
 
-#ifdef XFree86LOADER
-static const char *vbeSymbols[] = {
-    "VBEInit",
-    "vbeFree",
-    "vbeDoEDID",
-    NULL
-};
-#endif
-
 static const char *i2cSymbols[] = {
     "xf86CreateI2CBusRec",
     "xf86I2CBusInit",
@@ -166,27 +157,6 @@ static const char *int10Symbols[] = {
 };
 
 
-#ifdef XFree86LOADER
-
-static MODULESETUPPROTO(rivaSetup);
-
-static XF86ModuleVersionInfo rivaVersRec =
-{
-    "riva",
-    MODULEVENDORSTRING,
-    MODINFOSTRING1,
-    MODINFOSTRING2,
-    XORG_VERSION_CURRENT,
-    PACKAGE_VERSION_MAJOR, PACKAGE_VERSION_MINOR, PACKAGE_VERSION_PATCHLEVEL,
-    ABI_CLASS_VIDEODRV,                     /* This is a video driver */
-    ABI_VIDEODRV_VERSION,
-    MOD_CLASS_VIDEODRV,
-    {0,0,0,0}
-};
-
-_X_EXPORT XF86ModuleData riva128ModuleData = { &rivaVersRec, rivaSetup, NULL };
-#endif
-
 
 typedef enum {
     OPTION_SW_CURSOR,
@@ -253,30 +223,6 @@ RivaFreeRec(ScrnInfoPtr pScrn)
     pScrn->driverPrivate = NULL;
 }
 
-
-#ifdef XFree86LOADER
-
-static pointer
-rivaSetup(pointer module, pointer opts, int *errmaj, int *errmin)
-{
-    static Bool setupDone = FALSE;
-
-    /* This module should be loaded only once, but check to be sure. */
-
-    if (!setupDone) {
-        setupDone = TRUE;
-
-        LoaderRefSymLists(vgahwSymbols, xaaSymbols, fbSymbols,
-                          ramdacSymbols, shadowSymbols,
-                          i2cSymbols, ddcSymbols, vbeSymbols,
-                          fbdevHWSymbols, int10Symbols, NULL);
-    } 
-    return (pointer)1;
-}
-
-
-#endif /* XFree86LOADER */
-
 _X_EXPORT const OptionInfoRec *
 RivaAvailableOptions(int chipid, int busid)
 {
